/*
  Program: ei Sokobanu
  Autor: Radek Pelnek  
  Ploha ke knize "Jak to vyeit",  www.radekpelanek.cz/?jtv

  Zadn: Napite program, kter bude eit lohu Sokoban. Jako vstup dostane
  popis problmu ve standardnm formtu, jeho detailn popis lze nalzt na
  internetu, nap. na http://www.sokobano.de/wiki/index.php?title=Level_format

  Na vstup program zape een ve vlastnm zvolenm formtu (tak aby zpis
  byl snadno interpretovateln lovkem). Clem je program vyladit tak, aby
  zvldl vyeit co nejvt problmy. V ploze na webu jsou uveden zadn
  rzn velikosti, kter je mon pout pro otestovn programu, ppadn jako
  podklad pro srovnn nkolika soutcch program.

  Tento ukzkov program slou pedevm jako "benchmark" pro srovnn, nikoliv
  jako "ukzkov een" - program nen napsn pln ist (nkter neefektivnosti
  a neistoty jsou okomentovny).
*/

#include <iostream>
#include <vector>
#include <map>
#include <list>
#include <string>
#include <queue>
#include <sstream>


#define MAX 30  // maximln rozmry hernho plnu

using namespace std;


// Globln st informac o hernm plnu.
bool wall[MAX][MAX];    // poloha zd
bool goal[MAX][MAX];    // poloha clovch pol
int gdist[MAX][MAX];    // vzdlenost do nejbliho clovho pole - vyuv se v heuristice
int n, m, b = 0;        // rozmery plnu
                        // n = poet sloupc, m = poet dk, b = poet beden



/* Reprezentace "pohybliv" sti stavu. Rozliujeme stavy podle sel beden.
 To usnaduje manipulaci se stavy, ale je to redundantn a vede to k nrstu
 potu stav. Zmnou tto reprezentace je mono doshnout vraznho urychlen.
*/
struct state_t {
  int mx, my;          // pozice panka
  vector<int> bx, by;  // pozice beden
  int dist;            // heuristick hodnota - odhad vzdlenosti do cle
};

// kompartor stav, kter potebujeme, abychom mohli pouvat strukturu map
struct statecmpr {
  bool operator()(const state_t s1, const state_t s2) const
  {
    if (s1.mx < s2.mx) return true;
    if (s1.mx > s2.mx) return false;
    if (s1.my < s2.my) return true;
    if (s1.my > s2.my) return false;
    for (int i=0; i< b; i++) {
      if (s1.bx[i] < s2.bx[i]) return true;
      if (s1.bx[i] > s2.bx[i]) return false;
      if (s1.by[i] < s2.by[i]) return true;
      if (s1.by[i] > s2.by[i]) return false;
    }
    return false;
  }
};

// heuristick kompartor pro prioritn frontu 
struct heurcomparator
{
  bool operator()(const state_t s1, const state_t s2) const
  {
    return s1.dist > s2.dist;
  }
};

// identifikace tahu - kter bedna se pohne a jakm smrem
struct move_t {
  int b, dx, dy;
};

// iniciln stav si drme globln
state_t initstate;

// --------------------- potn heuristiky --------------------------

/* Napot pro kad polko vzdlenost k nejblimu clovmu poli.
 Funkce je neefektivn, ale to je pln jedno, protoe tento kd se
 spout jen jednou a pln je mal.
*/
void compute_global_dist() {
  for (int i=1; i<=n; i++)
    for (int j=1; j<=m; j++) 
      if (goal[i][j]) gdist[i][j] = 0; else gdist[i][j] = 1000;
  bool change = true;
  while (change) {
    change = false;
    for (int i=1; i<=n; i++)
      for (int j=1; j<=m; j++) 
	if (! wall[i][j]) {
	  if (gdist[i-1][j] + 1 < gdist[i][j]) { gdist[i][j] = gdist[i-1][j]+1; change = true; }
	  if (gdist[i+1][j] + 1 < gdist[i][j]) { gdist[i][j] = gdist[i+1][j]+1; change = true; }
	  if (gdist[i][j-1] + 1 < gdist[i][j]) { gdist[i][j] = gdist[i][j-1]+1; change = true; }
	  if (gdist[i][j+1] + 1 < gdist[i][j]) { gdist[i][j] = gdist[i][j+1]+1; change = true; }
	}
  }
}

// vypot heuristickou hodnotu pro aktuln stav - prost souet vzdlenost  pro jednotliv bedny
void compute_dist(state_t* s) {
  int d = 0;
  for (int i=0; i<b;i++) {
    d += gdist[(*s).bx[i]][(*s).by[i]];
  }
  (*s).dist = d;
}

// ------------------------ naten a vpis stavu ----------------------

// naten plnu a uloen do inicilnho stavu (kter je globln)
void read_plan() {
  list<int> lx, ly;

  cin >> n >> m;  
  for (int j=0; j <= m+1; j++) {
    string s;
    getline(cin, s);
    for (int i=0; i <= n+1; i++) {
      if (j==0 || i==0 || i==n+1 || j==m+1) { wall[i][j] = true; goal[i][j] = false; }
      else {
	char ch = s[i-1];
	//	cout << "ble: "<< i << " "<<j<< " "<< ch<<endl;
	switch (ch) {
	case '#':  wall[i][j] = true; goal[i][j] = false; break;
	case '.':  wall[i][j] = false; goal[i][j] = true; break;
	case '@':  wall[i][j] = false; goal[i][j] = false; initstate.mx = i; initstate.my = j; break;
	case '+':  wall[i][j] = false; goal[i][j] = true; initstate.mx = i; initstate.my = j; break;
	case '$':  wall[i][j] = false; goal[i][j] = false; b++; lx.push_back(i); ly.push_back(j); break;
	case '*':  wall[i][j] = false; goal[i][j] = true; b++; lx.push_back(i); ly.push_back(j); break;
	case ' ':  wall[i][j] = false; goal[i][j] = false; break;
	}	
      }      
    }
  }
  compute_global_dist();
  //  initstate.dist = 0; 
  initstate.bx.resize(b);
  initstate.by.resize(b);
  for (int i=0; i<b; i++) {
    initstate.bx[i] = lx.front(); lx.pop_front();
    initstate.by[i] = ly.front(); ly.pop_front();    
  }
  compute_dist(&initstate);
}

// vpis plnu
void print_plan() {
  for (int j=0; j <= m+1; j++) {
    for (int i=0; i <= n+1; i++) {
      cout << (wall[i][j]?'#':' ');
    }
    cout <<endl;
  }
  for (int i=0; i<b; i++) {
    cout << initstate.bx[i] << " "<<initstate.by[i]<<endl;
  }
}

// vpis stavu
void print_state(state_t* s) {
  cout << (*s).mx << " " << (*s).my << ";; ";
  for (int i=0; i<b; i++) {
    cout << (*s).bx[i] << ","<< (*s).by[i]<<"; ";
  }
  cout<<":"<<(*s).dist<<endl;
}


// -------------------- jednoduch manipulace se stavy ----------------

// test, zda je stav finln (vechny bedny na clovch polch)
bool is_final(state_t* s) {
  for (int i=0; i <b; i++) {
    if (! goal[(*s).bx[i]][(*s).by[i]]) return false;
  }  
  return true;
}

// test, zda stav je identick inicilnmu
bool is_init(state_t* s) {
  if ((*s).mx != initstate.mx || (*s).my != initstate.my) return false;
  for (int i=0; i<b; i++) {
    if ((*s).bx[i] != initstate.bx[i] || (*s).by[i] != initstate.by[i]) return false;
  }
  return true;
}

//zkoprovn stavu: olds do news
void copy_state(state_t* news, state_t* olds) { 
  (*news).mx = (*olds).mx;
  (*news).my = (*olds).my;
  (*news).dist = (*olds).dist;
  for(int i=0; i<b; i++) { //zkopirovat bedny
    (*news).bx[i] = (*olds).bx[i]; (*news).by[i] = (*olds).by[i];	
  }  
}

// -------------------- potn nslednk, kontrola deadlock ---------------

// nsledujc data se pouvaj v nkolika funkcch, udrovn
// informace v globlnch promnnch implementaci trochu zjednoduuje,
// ale ist to nen...
bool mreach[MAX][MAX];    // pozice, kam se me panek dostat
bool blocked[MAX][MAX];   // blokovan pozice

// napot mreach a blocked pro aktuln stav
void compute_blocked_mreach(state_t* s) {
  int i,j;
  for (j=0; j <= m+1; j++) {
    for (i=0; i <= n+1; i++) {
      mreach[i][j] = false;
      blocked[i][j] = wall[i][j];
    }
  }
  for (i=0; i<b; i++) { blocked[(*s).bx[i]][(*s).by[i]] = true; }
  mreach[(*s).mx][(*s).my] = true;
  bool done = false;
  while (!done) {
    done = true;
    for (j=1; j <= m; j++) {
      for (i=1; i <= n; i++) {
	if (mreach[i][j]) {
	  if (!blocked[i-1][j] && !mreach[i-1][j]) {mreach[i-1][j] = true; done =false; }
	  if (!blocked[i+1][j] && !mreach[i+1][j]) {mreach[i+1][j] = true; done =false; }
	  if (!blocked[i][j-1] && !mreach[i][j-1]) {mreach[i][j-1] = true; done =false; }
	  if (!blocked[i][j+1] && !mreach[i][j+1]) {mreach[i][j+1] = true; done =false; }
	}
      }
    }
  }
}

// ur, zda stav je neperspektivn (deadlock)
// pedpokld, e je napotan sprvn "blocked" pro aktuln stav
bool is_deadlock(state_t* s) {
  for (int i=0; i<b; i++) {
    int x = (*s).bx[i];
    int y = (*s).by[i];
    if (!goal[x][y]) {
      if (blocked[x-1][y] && blocked[x-1][y-1] && blocked[x][y-1]) return true;
      if (blocked[x+1][y] && blocked[x+1][y-1] && blocked[x][y-1]) return true;
      if (blocked[x-1][y] && blocked[x-1][y+1] && blocked[x][y+1]) return true;
      if (blocked[x+1][y] && blocked[x+1][y+1] && blocked[x][y+1]) return true;
    }
  }
  return false;
}

// ----------------------- vlastn prohledvn ----------------------------


// provd prohledvn pomoc A* algoritmu s vyuitm ve definovanch heuristik
void search() {
  map<state_t,bool,statecmpr> visited;  // mnoina navtvench stav
  priority_queue<state_t, vector<state_t>, heurcomparator> q; // prioritn fronta

  // Reprezentace cesty do stavu - pro kad stav si udrujeme cel popis cesty
  // do stavu jako string v struktue map. To je een, kter je implementan
  // jednoduch, ale nefektivn. Zmnou tto reprezentace by mlo bt mon program urychlit.
  map<state_t,string,statecmpr> path; 
  
  list<move_t> move;
  bool done=false;
  int states = 0;

  visited[initstate] = true;
  q.push(initstate);
  while (! done && ! q.empty()) {
    // vybereme prvn stav z prioritn fronty
    state_t s = q.top(); q.pop();    
    states++;
    // napotme "blocked" a "mreach" pro aktuln stav
    compute_blocked_mreach(&s);
    
    // napotme vechny mon tahy a ulome je do struktury move
    // pro kadou bednu zkusme pohnout na 4 strany,
    // aby to lo, mus  bt z jedn strany volno a z druh mus bt mreach (panek se tam me dostat)
    for (int i=0; i<b; i++) { 
      if (!blocked[s.bx[i]-1][s.by[i]] && mreach[s.bx[i]+1][s.by[i]]) {
	move_t mv; mv.b = i; mv.dx = -1; mv.dy = 0;
	move.push_back(mv);
      }
      if (!blocked[s.bx[i]+1][s.by[i]] && mreach[s.bx[i]-1][s.by[i]]) {
	move_t mv; mv.b = i; mv.dx = +1; mv.dy = 0;
	move.push_back(mv);
      }
      if (!blocked[s.bx[i]][s.by[i]-1] && mreach[s.bx[i]][s.by[i]+1]) {
	move_t mv; mv.b = i; mv.dx = 0; mv.dy = -1;
	move.push_back(mv);
      }
      if (!blocked[s.bx[i]][s.by[i]+1] && mreach[s.bx[i]][s.by[i]-1]) {
	move_t mv; mv.b = i; mv.dx = 0; mv.dy = +1;
	move.push_back(mv);
      }
    }
    // nyn napotme nslednka pro kad tah a vyhodnotme jej
    while (! move.empty() && ! done) {
      move_t mv = move.front(); move.pop_front();
      stringstream tmp;
      // textov popis tahu
      tmp << mv.b;
      if (mv.dx ==-1) tmp << "Z"; 
      if (mv.dx ==1) tmp << "V"; 
      if (mv.dy ==-1) tmp << "S"; 
      if (mv.dy ==1) tmp << "J"; 
      state_t succ;
      succ.bx.resize(b);
      succ.by.resize(b);
      copy_state(&succ, &s);
      // umsti panka na pozici pslun bedny
      succ.mx = succ.bx[mv.b];
      succ.my = succ.by[mv.b];
      // pohni pslunou bednou
      succ.bx[mv.b] = succ.bx[mv.b] + mv.dx;
      succ.by[mv.b] = succ.by[mv.b] + mv.dy;
      // "znormalizovat" pozici panacka - umstt do polohy co nejvce vlevo nahoru
      compute_blocked_mreach(&succ);
      bool found = false;
      for(int j=1; j <= m; j++) {
	for(int i=1; i <= n;i++) { if (! found && mreach[i][j]) { succ.mx = i; succ.my = j; found = true;  } }}
      // napotat heuristickou hodnotu stavu
      compute_dist(&succ);

      // pokud tento stav jet nebyl navtven
      if (!visited[succ]) {
	visited[succ] = true;
	path[succ] =   path[s] + tmp.str();
	if (is_final(&succ)) {
	  // clov stav, vypsat cestu
	  cout << "cesta:"<< path[succ] <<endl;
	  done = true;
	} else {
	  if (! is_deadlock(&succ)) {
	    // vloit no fronty
	    q.push(succ);
	  }
	}	
      } 
    }
  }
  cout << "pocet navstivenych stavu:"<<states<<endl;
}

int main() {
  read_plan();
  search();
  return 0;
}
